/*
   Name: Periodic.
   Author: C. Chandre.
   Comments? chandre@cpt.univ-mrs.fr
   Description: This program finds the periodic orits of the forced pendulum model
    by multiple shooting method.
   Date: August 15, 2002.
   Copyright: This program is distributed under the GNU General Public License.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "nrutil.h"
#include "nrutil.c"
#define TINY 1.0e-20;
#define M_TWOPI 6.283185307179586476925286

double eps=0.02758,h;
int N=2000;

void identite(double J[4][4])
{
    J[1][1]=1.; J[1][2]=0.; J[1][3]=0.;
    J[2][1]=0.; J[2][2]=1.; J[2][3]=0.;
    J[3][1]=0.; J[3][2]=0.; J[3][3]=1.;
}

void annulmat(double **A,int n)
{
    int i,j;
    for (i=1;i<=n;i++)
        for(j=1;j<=n;j++)
            A[i][j]=0.0;
}

void annulvec(double *b, int n)
{
    int i;
    for (i=1;i<=n;i++)
        b[i]=0.0;
}

void mult(double J1[4][4], double J2[4][4], double J3[4][4])
{
    int i,j,k;
    for (i=1;i<=3;i++)
        for (j=1;j<=3;j++){
            J3[i][j]=0.;
            for (k=1;k<=3;k++)
                J3[i][j]+=J1[i][k]*J2[k][j];
        }
}

void egal(double J1[4][4], double J2[4][4])
{
    int i,j;
    for (i=1;i<=3;i++)
        for(j=1;j<=3;j++)
            J1[i][j]=J2[i][j];
}

void ludcmp(double **a, int n, int *indx, double *d)
{
    int i,imax,j,k;
    double big,dum,sum,temp;
    double *vv; 
    vv=dvector(1,n);
    *d=1.0; 
    for (i=1;i<=n;i++) {
        big=0.0;
        for (j=1;j<=n;j++)
            if ((temp=fabs(a[i][j])) > big) big=temp;
        if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
        vv[i]=1.0/big; 
    }
    for (j=1;j<=n;j++) { 
        for (i=1;i<j;i++) { 
            sum=a[i][j];
            for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
        }
        big=0.0; 
        for (i=j;i<=n;i++) {
            sum=a[i][j];
            for (k=1;k<j;k++)
                sum -= a[i][k]*a[k][j];
            a[i][j]=sum;
            if ( (dum=vv[i]*fabs(sum)) >= big) {
                big=dum;
                imax=i;
            }
        }
        if (j != imax) { 
            for (k=1;k<=n;k++) { 
                dum=a[imax][k];
                a[imax][k]=a[j][k];
                a[j][k]=dum;
            }
            *d = -(*d);
            vv[imax]=vv[j]; 
        }
        indx[j]=imax;
        if (a[j][j] == 0.0) a[j][j]=TINY;
        if (j != n) { 
            dum=1.0/(a[j][j]);
            for (i=j+1;i<=n;i++) a[i][j] *= dum;
        }
    } 
free_dvector(vv,1,n);
}

void lubksb(double **a, int n, int *indx, double *b)
{
    int i,ii=0,ip,j;
    double sum;
    for (i=1;i<=n;i++) { 
        ip=indx[i];
        sum=b[ip];
        b[ip]=b[i];
        if (ii)
            for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
        else if (sum) ii=i;
        b[i]=sum;
    }
    for (i=n;i>=1;i--) { 
        sum=b[i];
        for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
        b[i]=sum/a[i][i]; 
    } 
}

void rk4(double x[4], double h, double xout[4])
{
    int i;
    double hh,h6;
    double dxm[4],dxt[4],xt[4],dx[4];
    hh=h*0.5;
    h6=h/6.0;
    flow(x,dx);
    for (i=1;i<=3;i++) xt[i]=x[i]+hh*dx[i];
    flow(xt,dxt); 
    for (i=1;i<=3;i++) xt[i]=x[i]+hh*dxt[i];
    flow(xt,dxm); 
    for (i=1;i<=3;i++) {
        xt[i]=x[i]+h*dxm[i];
        dxm[i] += dxt[i];
    }
    flow(xt,dxt); 
    for (i=1;i<=3;i++)
        xout[i]=x[i]+h6*(dx[i]+dxt[i]+2.0*dxm[i]);
    xout[1]=fmod(xout[1]+M_PI,M_TWOPI)-M_PI;
}

void flow(double x[4], double dx[4])
{
    dx[1]=x[2];
    dx[2]=-eps*(sin(x[1])+sin(x[1]-x[3]));
    dx[3]=1.;
}

void pmap(double x[4], double fx[4], double Jx[4][4])
{
    int i,j,k;
    double A[4][4],xt[4],Jxt[4][4];
    xt[1]=x[1]; xt[2]=x[2]; xt[3]=x[3];
    identite(Jx);
    for (i=1;i<=N;i++){
        rk4(xt,h,fx);
        for (j=1;j<=3;j++)
            xt[j]=fx[j];
        A[1][1]=1.; A[1][2]=h; A[1][3]=0.;
        A[2][1]=-eps*h*(cos(xt[1])+cos(xt[1]-xt[3]));
        A[2][2]=1.;
        A[2][3]=eps*h*cos(xt[1]-xt[3]);
        A[3][1]=0.; A[3][2]=0.; A[3][3]=1.;
        mult(A,Jx,Jxt);
        egal(Jx,Jxt);
    }
    fx[3]=0.;
}
    

main()
{
    int i,j,ji,P,Q;
    double x[4],fx[4],Jx[4][4],J[4][4],Jtemp[4][4],**A,*b;
    double *X1,*X2,*X3,*FX1,*FX2,*FX3;
    double *JX11,*JX12,*JX13,*JX21,*JX22,*JX23,*JX31,*JX32,*JX33;
    double d,delta;
    int *indx;
    FILE *fic;
    h=M_TWOPI/((double) N);
     /* determination of periodic orbit P:Q */
        P=89; Q=144;
        X1=dvector(1,Q);
        X2=dvector(1,Q);
        X3=dvector(1,Q);
        FX1=dvector(1,Q);
        FX2=dvector(1,Q);
        FX3=dvector(1,Q);
        JX11=dvector(1,Q);
        JX12=dvector(1,Q);
        JX13=dvector(1,Q);
        JX21=dvector(1,Q);
        JX22=dvector(1,Q);
        JX23=dvector(1,Q);
        JX31=dvector(1,Q);
        JX32=dvector(1,Q);
        JX33=dvector(1,Q);
        X1[1]=0.;
        X2[1]=M_TWOPI*((double) P)/((double) Q);
        X3[1]=0.;
        for (i=1;i<=Q-1;i++){
            x[1]=X1[i]; x[2]=X2[i]; x[3]=X3[i];
            pmap(x,fx,Jx);
            X1[i+1]=fx[1];
            X2[i+1]=fx[2];
            X3[i+1]=fx[3];
        }
        A=dmatrix(1,4*Q,1,4*Q);
        b=dvector(1,4*Q);
        indx=ivector(1,4*Q);
        do{
            for (i=1;i<=Q;i++)
            {
                x[1]=X1[i]; x[2]=X2[i]; x[3]=X3[i];
                pmap(x,fx,Jx);
                FX1[i]=fx[1]; FX2[i]=fx[2]; FX3[i]=fx[3];
                JX11[i]=Jx[1][1]; JX12[i]=Jx[1][2]; JX13[i]=Jx[1][3];
                JX21[i]=Jx[2][1]; JX22[i]=Jx[2][2]; JX23[i]=Jx[2][3];
                JX31[i]=Jx[3][1]; JX32[i]=Jx[3][2]; JX33[i]=Jx[3][3];
            }
            annulmat(A,4*Q);
            annulvec(b,4*Q);
            for (i=1;i<=3*Q;i++) A[i][i]=1.0;
            for (i=1;i<=Q-1;i++){
                A[3*i+1][3*i-2]+=-JX11[i];
                A[3*i+1][3*i-1]+=-JX12[i];
                A[3*i+1][3*i]+=-JX13[i];
                A[3*i+2][3*i-2]+=-JX21[i];
                A[3*i+2][3*i-1]+=-JX22[i];
                A[3*i+2][3*i]+=-JX23[i];
                A[3*i+3][3*i-2]+=-JX31[i];
                A[3*i+3][3*i-1]+=-JX32[i];
                A[3*i+3][3*i]+=-JX33[i];
            }
            A[1][3*Q-2]+=-JX11[Q];
            A[1][3*Q-1]+=-JX12[Q];
            A[1][3*Q]+=-JX13[Q];
            A[2][3*Q-2]+=-JX21[Q];
            A[2][3*Q-1]+=-JX22[Q];
            A[2][3*Q]+=-JX23[Q];
            A[3][3*Q-2]+=-JX31[Q];
            A[3][3*Q-1]+=-JX32[Q];
            A[3][3*Q]+=-JX33[Q];
            for (i=1;i<=Q;i++){
                A[3*Q+i][3*i]=1.;
                A[3*i-2][3*Q+i]=X2[i];
                A[3*i-1][3*Q+i]=-eps*(sin(X1[i])+sin(X1[i]-X3[i]));
                A[3*i][3*Q+i]=1.;
            }
            for (i=2;i<=Q;i++){
                b[3*i-2]=FX1[i-1]-X1[i];
                b[3*i-1]=FX2[i-1]-X2[i];
            }
            b[1]=FX1[Q]-X1[1];
            b[2]=FX2[Q]-X2[1];
            ludcmp(A,4*Q,indx,&d);
            lubksb(A,4*Q,indx,b);
            delta=0.;
            for (i=1;i<=Q;i++){
                X1[i]+=b[3*i-2];
                X2[i]+=b[3*i-1];
                X3[i]+=b[3*i];
                delta+=fabs(b[3*i-2])+fabs(b[3*i-1])+fabs(b[3*i]);
                X1[i]-=M_TWOPI*floor((X1[i]+M_PI)/M_TWOPI);
            }
            printf("%le %le delta= %le\n",X1[1],X2[1],delta);
        }while (delta>=1e-5 && delta<=1e+5);
        if (delta>=1e+5) printf("failure...\n");
        if (delta<=1e-5){
            fic=fopen("data.dat","w+");
            for(i=1;i<=Q;i++)
                fprintf(fic,"%le %le\n",X1[i],X2[i]);
            fclose(fic);
            J[1][1]=JX11[1]; J[1][2]=JX12[1]; J[1][3]=JX13[1];
            J[2][1]=JX21[1]; J[2][2]=JX22[1]; J[2][3]=JX23[1];
            J[3][1]=JX31[1]; J[3][2]=JX32[1]; J[3][3]=JX33[1];
            for (i=2;i<=Q;i++){
                Jx[1][1]=JX11[i]; Jx[1][2]=JX12[i]; Jx[1][3]=JX13[i];
                Jx[2][1]=JX21[i]; Jx[2][2]=JX22[i]; Jx[2][3]=JX23[i];
                Jx[3][1]=JX31[i]; Jx[3][2]=JX32[i]; Jx[3][3]=JX33[i];
                mult(Jx,J,Jtemp);
                egal(J,Jtemp);
            }
            printf("trace = %le\n",J[1][1]+J[2][2]+J[3][3]);
        }
        free_dmatrix(A,1,4*Q,1,4*Q);
        free_dvector(b,1,4*Q);
        free_dvector(X1,1,4*Q);
        free_dvector(X2,1,4*Q);
        free_dvector(X3,1,4*Q);
        free_dvector(FX1,1,4*Q);
        free_dvector(FX2,1,4*Q);
        free_dvector(FX3,1,4*Q);
        free_dvector(JX11,1,4*Q);
        free_dvector(JX12,1,4*Q);
        free_dvector(JX13,1,4*Q);
        free_dvector(JX21,1,4*Q);
        free_dvector(JX22,1,4*Q);
        free_dvector(JX23,1,4*Q);
        free_dvector(JX31,1,4*Q);
        free_dvector(JX32,1,4*Q);
        free_dvector(JX33,1,4*Q);
        free_ivector(indx,1,4*Q);
}
